/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System - Generic Mail Gateway.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2002 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.mailgate;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import javax.mail.*;
import javax.mail.internet.*;
import com.ibm.bsf.*;

public class Main
{
  /*--------------------------------------------------------------------------------
   * Internal library class
   *--------------------------------------------------------------------------------
   */

  public class MyLibrary extends Library
  {
    MyLibrary()
    {
      super();

    } // end constructor

    public final String getConfigParam(String name)
    {
      return props.getProperty("mailgate.script.config." + name);

    } // end getConfigParam

  } // end class MyLibrary

  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  // Exit codes interpreted by qmail-local delivery agent - see qmail-command(8) man page
  public static final int EXIT_SUCCESS = 0;
  public static final int EXIT_SUCCESS_FINAL = 99;
  public static final int EXIT_FAIL_PERMANENT = 100;
  public static final int EXIT_FAIL_TEMPORARY = 111;

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private Properties props;           // properties loaded from the command line
  private File script_file;           // script file to be run
  private StringBuffer script_code;   // script code from above
  private Session session;            // mail session object
  private BSFManager script_mgr;      // script manager
  private String script_lang;         // language the script is written in

  /*--------------------------------------------------------------------------------
   * Constructor
   *--------------------------------------------------------------------------------
   */

  private Main(String propfile) throws PermanentExit, TemporaryExit
  {
    File f = new File(propfile);
    if (f.canRead())
    { // load up the file
      try
      { // load the properties from the file
	FileInputStream stm = new FileInputStream(f);
	props = new Properties();
	props.load(stm);

      } // end try
      catch (IOException e)
      { // unable to load the system properties!
	throw new TemporaryExit("error loading properties from \"" + f.getAbsolutePath() + "\"",e);

      } // end catch

    } // end if
    else // the file name is bogus
      throw new PermanentExit("properties file \"" + f.getAbsolutePath() + "\" cannot be read");

    // Check the script file name.
    String tmp = props.getProperty("mailgate.script");
    if (tmp!=null)
    { // see if the script file can be read
      script_file = new File(tmp);
      if (!(script_file.canRead()))
	throw new PermanentExit("script file \"" + script_file.getAbsolutePath() + "\" cannot be read");

    } // end if
    else // script file not found
      throw new PermanentExit("no script file specified in properties");

    // Preload the script code.
    try
    { // copy the script code to a string buffer
      FileReader rdr = new FileReader(script_file);
      StringWriter wr = new StringWriter();
      char[] buf = new char[4096];
      int rd = rdr.read(buf);
      while (rd>=0)
      { // simple read-write loop
	if (rd>0)
	  wr.write(buf,0,rd);
	rd = rdr.read(buf);

      } // end while

      script_code = wr.getBuffer();  // load the script code

    } // end try
    catch (IOException e)
    { // unable to load the buffer
      throw new PermanentExit("unable to load script file \"" + script_file.getAbsolutePath() + "\"");

    } // end catch

    // Load the temporary script directory.
    tmp = props.getProperty("mailgate.script.tmpdir");
    if (tmp!=null)
    { // Temporary directory specified...
      File tmpf = new File(tmp);
      if (!(tmpf.isDirectory()))
	throw new PermanentExit("temporary dir \"" + tmpf.getAbsolutePath() + "\" does not exist");
      tmp = tmpf.getAbsolutePath();

    } // end if
    else // directory name not found
      throw new PermanentExit("no temporary directory specified in properties");
    
    // Create the BSF manager.
    script_mgr = new BSFManager();
    script_mgr.setTempDir(tmp);

    try
    { // declare beans used by the scripting engines
      script_mgr.declareBean("lib",new MyLibrary(),MyLibrary.class);

    } // end try
    catch (BSFException e)
    { // translate the exception
      throw new PermanentExit("declaration error in BSF",e);

    } // end catch

    if (BSFManager.isLanguageRegistered("javascript"))
    { // we need to so some fixing up to make Rhino work
      try
      { // ensure the engine is loaded, and do our fixup routine
	script_mgr.loadScriptingEngine("javascript");
	fixupRhino();

      } // end if
      catch (BSFException e)
      { // do nothing - we don't have JavaScript loaded
      } // end catch

    } // end if

    try
    { // look up the scripting language for this file
      script_lang = script_mgr.getLangFromFilename(script_file.getAbsolutePath());

    } // end try
    catch (BSFException e)
    { // this may be thrown by getLangFromFilename
      throw new PermanentExit("unknown scripting language: file " + script_file.getAbsolutePath());

    } // end catch

    // Get the E-mail session object.
    session = Session.getDefaultInstance(props);

  } // end constructor

  /*--------------------------------------------------------------------------------
   * Internal operations
   *--------------------------------------------------------------------------------
   */

  private static void fixupRhino()
  {
    // What fixupRhino does is to turn off the fancy caching in the JavaScript engine.  This is because
    // caching involves dynamic creation of classes on the fly, and, for some reason (probably because
    // of differing ClassLoader implementations), that doesn't work, whether Rhino is loaded in the
    // Web application or in the JRE extensions.  When caching is turned off, Rhino can be loaded as
    // part of the web application, and it can see all the standard classes and all the Web application
    // classes.  But we have to turn it off by manipulating it via the Reflection API, as, for all we know,
    // Rhino isn't even being loaded right now.  (Though we have BSF load Rhino right away if it can,
    // to try and get this straightened out.)
    try
    { // get a pointer to the Context class
      Class klass = Class.forName("org.mozilla.javascript.Context");

      // get a pointer to its setCachingEnabled method
      final Class[] parmtypes = { Boolean.TYPE };
      Method m = klass.getMethod("setCachingEnabled",parmtypes);

      // invoke it to turn caching off!
      final Object[] parms = { Boolean.FALSE };
      m.invoke(null,parms);

    } // end try
    catch (Exception e)
    { // just ignore me
    } // end catch

  } // end fixupRhino

  private void runMe() throws PermanentExit, TemporaryExit
  {
    MessageData msg;
    try
    { // pull the E-mail message from the input and parse it
      MimeMessage in_msg = new MimeMessage(session,System.in);
      msg = new MessageData(in_msg);

    } // end try
    catch (MessagingException me)
    { // error loading the E-mail message
      throw new PermanentExit("error loading incoming message",me);

    } // end catch
    catch (IOException ioe)
    { // error reading the message somehow
      throw new TemporaryExit("error reading incoming message",ioe);

    } // end catch

    // Now process the message via the script.
    script_mgr.registerBean("message",msg);
    try
    { // exec the script!
      script_mgr.exec(script_lang,script_file.getAbsolutePath(),1,1,script_code);

    } // end try
    catch (BSFException e)
    { // handle the exception somehow
      Throwable t = e.getTargetException();
      if (t instanceof TemporaryExit)
	throw (TemporaryExit)t;
      if (t instanceof PermanentExit)
	throw (PermanentExit)t;
      throw new TemporaryExit("script error",e);

    } // end catch

  } // end runMe

  /*--------------------------------------------------------------------------------
   * Ye Olde main() Function
   *--------------------------------------------------------------------------------
   */

  public static void main(String[] args)
  {
    // Should be invoked from within a .qmail file as:
    // |preline -f <this executable>
    if (args.length<1)
    { // bad error - bail out
      System.err.println("usage: java " + Main.class.getName() + " property-file");
      System.exit(EXIT_FAIL_PERMANENT);

    } // end if

    int rc = EXIT_SUCCESS;
    try
    { // create the "self" object
      Main self = new Main(args[0]);
      self.runMe();

    } // end try
    catch (TemporaryExit te)
    { // temporary failure
      System.err.println("Temporary failure in mail gateway");
      te.printStackTrace();
      rc = EXIT_FAIL_TEMPORARY;

    } // end catch
    catch (PermanentExit pe)
    { // permanent failure
      System.err.println("Permanent failure in mail gateway");
      pe.printStackTrace();
      rc = EXIT_FAIL_PERMANENT;

    } // end catch

    System.exit(rc);

  } // end main

} // end class Main


